from grawlix import logging
from grawlix.book import Result, Book, Metadata, OnlineFile, ImageList, Series
from grawlix.encryption import Encryption
from grawlix.exceptions import InvalidUrl, AccessDenied
from .source import Source

import re
from typing import Tuple, List
from hashlib import sha256
from Crypto.Cipher import AES

class DcUniverseInfinite(Source):
    name = "DC Universe Infinite"
    match: list[str] = [
        # Reader page
        r"https://www.dcuniverseinfinite.com/comics/book/[^/]+/[^/]+/c/reader",
        # Issue info page
        r"https://www.dcuniverseinfinite.com/comics/book/[^/]+/[^/]+/c",
        # Series info page
        r"https://www.dcuniverseinfinite.com/comics/series/[^/]+/[^/]+"
    ]
    _authentication_methods = [ "cookies" ]


    async def download(self, url: str) -> Result:
        # Set headers
        auth_token = self._client.cookies.get("session")
        self._client.headers.update({
            "Authorization": f"Token {auth_token}",
            "X-Consumer-Key": await self.download_consumer_secret()
        })
        self.plan = await self.download_plan()
        logging.debug(f"{self.plan=}")
        # Download book
        typ, id = self.extract_id_from_url(url)
        if typ == "book":
            logging.debug(f"Book id: {id}")
            return await self.download_book_from_id(id)
        else:
            logging.debug(f"Series id: {id}")
            return await self.download_series(id)


    async def download_series(self, series_id: str) -> Series[str]:
        # TODO Check for ultra releases
        response = await self._client.get(
            f"https://www.dcuniverseinfinite.com/api/comics/1/series/{series_id}/?trans=en"
        )
        content = response.json()
        return Series(
            title = content["title"],
            book_ids = [x for x in content["book_uuids"]["issue"]]
        )


    async def download_book_from_id(self, book_id: str) -> Book:
        return Book(
            data = await self.download_pages(book_id),
            metadata = await self.download_book_metadata(book_id)
        )


    async def download_pages(self, book_id: str) -> ImageList:
        """
        Download comic pages

        :param book_id: Id of comic
        :return: List of comic pages
        """
        response = await self._client.get(
            f"https://www.dcuniverseinfinite.com/api/5/1/rights/comic/{book_id}?trans=en"
        )
        jwt = response.json()
        response = await self._client.get(
            "https://www.dcuniverseinfinite.com/api/comics/1/book/download/?page=1&quality=HD&trans=en",
            headers = {
                "X-Auth-JWT": jwt
            }
        )
        response = response.json()
        if not "uuid" in response:
            raise AccessDenied
        uuid = response["uuid"]
        job_id = response["job_id"]
        format_id = response["format"]
        images: List[OnlineFile] = []
        for page in response["images"]:
            page_number = page["page_number"]
            images.append(OnlineFile(
                url = page["signed_url"],
                extension = "jpg",
                encryption = DcUniverseInfiniteEncryption(uuid, page_number, job_id, format_id)
            ))
        return ImageList(images)


    async def download_book_metadata(self, book_id: str) -> Metadata:
        """
        Download book metadata

        :param book_id: Id of book
        :return: Book metadata
        """
        response = await self._client.get(
            f"https://www.dcuniverseinfinite.com/api/comics/1/book/{book_id}/?trans=en"
        )
        content = response.json()
        return Metadata(
            title = content["title"],
            series = content["series_title"],
            index = int(content["issue_number"]),
            publisher = "DC"
        )


    def extract_id_from_url(self, url: str) -> Tuple[str, str]:
        """
        Extract book or series id from url

        :param url: Url of page with id
        :return: Type (book or series) and id
        """
        match_index = self.get_match_index(url)
        if match_index == 0:
            book_id = url.split("/")[-3]
            return ("book", book_id)
        if match_index == 1:
            book_id = url.split("/")[-2]
            return ("book", book_id)
        if match_index == 2:
            series_id = url.split("/")[-1]
            return ("series", series_id)
        raise InvalidUrl


    async def download_consumer_secret(self) -> str:
        """Download consumer secret"""
        response = await self._client.get(
            "https://www.dcuniverseinfinite.com/api/5/consumer/www?trans=en"
        )
        return response.json()["consumer_secret"]


    async def download_plan(self) -> str:
        """Download user subscription plan"""
        response = await self._client.get(
            "https://www.dcuniverseinfinite.com/api/claims/?trans=en"
        )
        return response.json()["data"]["urn:df:clm:premium"]["plan"]


class DcUniverseInfiniteEncryption:
    key: bytes

    def __init__(self, uuid: str, page_number: int, job_id: str, format_id: str):
        string_key = f"{uuid}{page_number}{job_id}{format_id}"
        self.key = sha256(string_key.encode("utf8")).digest()


    def decrypt(self, data: bytes) -> bytes:
        # The first 8 bytes contains the size of the output file
        original_size = int.from_bytes(data[0:8], byteorder="little")
        # The next 16 bytes are the initialization vector
        iv = data[8:24]
        # The rest of the data is the encrypted image
        encrypted_image = data[24:]
        # Decrypting image
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return cipher.decrypt(encrypted_image)



from .source import Source
from grawlix.book import Book, HtmlFile, HtmlFiles, OnlineFile, Metadata

from bs4 import BeautifulSoup

USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; rv:113.0) Gecko/20100101 Firefox/113.0"

class FanfictionNet(Source):
    name: str = "fanfiction.net"
    match = [
        r"https://www.fanfiction.net/s/\d+/\d+.*"
    ]
    _authentication_methods: list[str] = [ "cookies" ]

    async def download(self, url: str) -> Book:
        book_id = self._extract_id(url)
        response = await self._client.get(
            f"https://www.fanfiction.net/s/{book_id}/1",
            headers = {
                "User-Agent": USER_AGENT
            }
        )
        soup = BeautifulSoup(response.text, "lxml")
        chapters = []
        for index, chapter in enumerate(soup.find(id="chap_select").find_all("option")):
            chapters.append(
                HtmlFile(
                    title = chapter.text,
                    file = OnlineFile(
                        url = f"https://www.fanfiction.net/s/{book_id}/{index+1}",
                        extension = "html",
                        headers = {
                            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; rv:113.0) Gecko/20100101 Firefox/113.0",
                        },
                        cookies = self._client.cookies
                    ),
                    selector = { "id": "storytext" }
                )
            )
        return Book(
            data = HtmlFiles(htmlfiles = chapters),
            metadata = Metadata(
                title = soup.find("b", class_="xcontrast_txt").text,
            )
        )

    @staticmethod
    def _extract_id(url: str) -> str:
        """
        Extracts book id from url

        :param url: Url of book
        :returns: Id of book
        """
        return url.split("/")[4]



from grawlix.book import Metadata

import xml.etree.ElementTree as ET
from typing import Optional


def add_value(element: ET.Element, name: str, value: Optional[str]) -> None:
    """
    Add new tag to element

    :param element: Element to add tag to
    :param name: Name of new tag
    :param value: Contents of new tag
    """
    if value:
        subelement = ET.SubElement(element, name)
        subelement.text = str(value)


def to_comic_info(metadata: Metadata) -> str:
    """
    Output as ComicRack metadata format (ComicInfo)
    Documentation: https://anansi-project.github.io/docs/comicinfo

    :param metadata: Book metadata
    :returns: ComicInfo xml as a string
    """
    root = ET.Element("ComicInfo")
    add_value(root, "Title", metadata.title)
    add_value(root, "Series", metadata.series)
    add_value(root, "Summary", metadata.description)
    add_value(root, "Publisher", metadata.publisher)
    add_value(root, "Year", getattr(metadata.release_date, "year", None))
    add_value(root, "Month", getattr(metadata.release_date, "month", None))
    add_value(root, "Day", getattr(metadata.release_date, "day", None))
    add_value(root, "Format", "Web")
    return ET.tostring(root).decode("utf8")



from dataclasses import dataclass
from typing import Optional
import tomli
import appdirs
import os


@dataclass(slots=True)
class SourceConfig:
    """Stores configuration for source"""
    username: Optional[str]
    password: Optional[str]
    library: Optional[str]


@dataclass(slots=True)
class Config:
    """Grawlix configuration"""
    sources: dict[str, SourceConfig]


def load_config() -> Config:
    """
    Load config from disk

    :returns: Config object
    """
    config_dir = appdirs.user_config_dir("grawlix", "jo1gi")
    config_file = os.path.join(config_dir, "grawlix.toml")
    if os.path.exists(config_file):
        with open(config_file, "rb") as f:
            config_dict = tomli.load(f)
    else:
        config_dict = {}
    sources = {}
    if "sources" in config_dict:
        for key, values in config_dict["sources"].items():
            sources[key] = SourceConfig (
                username = values.get("username"),
                password = values.get("password"),
                library = values.get("library"),
            )
    return Config(sources)



from grawlix.book import Book, SingleFile, Metadata, OfflineFile
from grawlix.exceptions import DataNotFound
from grawlix import logging
from .source import Source

import random
import string
from bs4 import BeautifulSoup
import asyncio
import json

class InternetArchive(Source):
    name: str = "Internet Archive"
    match: list[str] = [
        r"https://archive.org/details/.+"
    ]
    _authentication_methods = [ "login", "cookies" ]
    _login_credentials = [ "username", "password" ]

    @staticmethod
    def _format_data(content_type: str, fields):
        data = ""
        for name, value in fields.items():
            data += f"--{content_type}\x0d\x0aContent-Disposition: form-data; name=\"{name}\"\x0d\x0a\x0d\x0a{value}\x0d\x0a"
        data += content_type+"--"
        return data


    async def login(self, username: str, password: str, **kwargs) -> None:
        await self._client.get("https://archive.org/account/login")
        content_type = "----WebKitFormBoundary"+"".join(random.sample(string.ascii_letters + string.digits, 16))
        headers = {'Content-Type': 'multipart/form-data; boundary='+content_type}
        data = self._format_data(content_type, {"username":username, "password":password, "submit_by_js":"true"})
        response = await self._client.post(
            "https://archive.org/account/login",
            data=data,
            headers=headers
        )
        if not "Successful login" in response.text:
            print("Failed login")
            exit(1)


    async def _download_acsm(self, book_id: str) -> bytes:
        """
        Loan book on archive.org and download acsm file

        :param book_id: Id of book
        """
        await self._client.post(
            "https://archive.org/services/loans/loan/searchInside.php",
            data = {
                "action": "grant_access",
                "identifier": book_id
            }
        )
        await self._client.post(
            "https://archive.org/services/loans/loan/",
            data = {
                "action": "browse_book",
                "identifier": book_id
            }
        )
        # TODO: Error handling
        await self._client.post(
            "https://archive.org/services/loans/loan/",
            data = {
                "action": "create_token",
                "identifier": book_id
            }
        )
        acsm_response = await self._client.get(
            f"https://archive.org/services/loans/loan/?action=media_url&identifier={book_id}&format=pdf&redirect=1",
            follow_redirects = True
        )
        return acsm_response.content


    async def download(self, url: str) -> Book:
        book_id = url.split("/")[4]
        metadata, acsm_file = await asyncio.gather(
            self._download_metadata(book_id),
            self._download_acsm(book_id)
        )
        return Book(
            data = SingleFile(
                OfflineFile(
                    content = acsm_file,
                    extension = "acsm",
                )
            ),
            metadata = Metadata(
                title = metadata["title"],
                authors = [ metadata.get("creator") ] if "creator" in metadata else []
            )
        )


    async def _download_metadata(self, book_id: str) -> dict:
        """
        Download metadata for book

        :param book_id: Id of book
        :returns: Dictionary with metadata
        """
        page_response = await self._client.get(
            f"https://archive.org/details/{book_id}"
        )
        soup = BeautifulSoup(page_response.text, "lxml")
        reader_data = json.loads(soup.find(class_="js-bookreader").get("value"))
        metadata_url = f"https:{reader_data['url']}"
        logging.debug(f"{metadata_url=}")
        metadata_response = await self._client.get(
            metadata_url
        )
        return metadata_response.json()["data"]["metadata"]



from grawlix.book import Book, Series, Result

from typing import Generic, TypeVar, Tuple, Optional
from http.cookiejar import MozillaCookieJar
import re
from typing import Generic, TypeVar, Tuple
import httpx

T = TypeVar("T")

class Source(Generic[T]):
    """
    General class for downloading books from various sources
    """

    name: str = "UNKNOWN"
    match: list[str] = []
    _authentication_methods: list[str] = []
    _login_credentials = [ "username", "password" ]
    authenticated = False

    def __init__(self):
        self._client = httpx.AsyncClient()


    @property
    def requires_authentication(self) -> bool:
        """Does the source require authentication to download books"""
        return len(self._authentication_methods) > 0


    @property
    def supports_login(self) -> bool:
        """Does the source support authentication with username and password"""
        return "login" in self._authentication_methods


    async def login(self, username: str, password: str, **kwargs: str):
        """
        Login to source

        :param username: Username of user for source
        :param password: Password of user for source
        """
        raise NotImplementedError


    @property
    def supports_cookies(self) -> bool:
        """Does the source support authentication with cookie file"""
        return "cookies" in self._authentication_methods


    def load_cookies(self, cookie_file: str):
        """
        Authenticate with source with netscape cookie file

        :param cookie_file: Path to netscape cookie file
        """
        if self.supports_cookies:
            cookie_jar = MozillaCookieJar()
            cookie_jar.load(cookie_file, ignore_expires=True)
            self._client.cookies.update(cookie_jar)
            self.authenticated = True


    async def download(self, url: str) -> Result[T]:
        """
        Download book metadata from source

        :param url: Url of book to download
        :returns: Book metadata
        """
        raise NotImplementedError


    async def download_book_from_id(self, book_id: T) -> Book:
        """
        Download book from id

        :param book_id: Internal id of book
        :returns: Downloaded book metadata
        """
        raise NotImplementedError


    def get_match_index(self, url: str) -> Optional[int]:
        """
        Find the first regex in `self.match` that matches url

        :param url: Url to match
        :returns: Index of regex
        """
        for index, match in enumerate(self.match):
            if re.match(match, url):
                return index
        return None



from grawlix.book import Book, Metadata, ImageList, OnlineFile, Series, Result
from grawlix.exceptions import InvalidUrl, DataNotFound
from grawlix import logging

from .source import Source

import re
from datetime import date

# Personal marvel ip key
API_KEY = "83ac0da31d3f6801f2c73c7e07ad76e8"

class Marvel(Source[str]):
    name: str = "Marvel"
    match = [
        r"https://www.marvel.com/comics/issue/\d+/.+",
        r"https://read.marvel.com/#/book/\d+",
        r"https://www.marvel.com/comics/series/\d+/.+"
    ]
    _authentication_methods: list[str] = [ "cookies" ]


    async def download(self, url: str) -> Result[str]:
        match_index = self.get_match_index(url)
        if match_index == 0:
            issue_id = await self._get_issue_id(url)
            return await self.download_book_from_id(issue_id)
        if match_index == 1:
            issue_id = url.split("/")[-1]
            return await self.download_book_from_id(issue_id)
        if match_index == 2:
            return await self._download_series(url)
        raise InvalidUrl


    async def _download_series(self, url: str) -> Series[str]:
        """
        Download series

        :param url: Url of series
        :returns: Series data
        """
        series_id = url.split("/")[-2]
        issue_ids = await self._download_issue_ids(series_id)
        metadata = await self._download_series_metadata(series_id)
        return Series(
            title = metadata["data"]["results"][0]["title"],
            book_ids = issue_ids
        )


    async def _download_issue_ids(self, series_id: str) -> list[str]:
        """
        Download issue ids from series

        :param series_id: Id of comic series on marvel.com
        :returns: List of comic ids for marvel comics
        """
        response = await self._client.get(
            f"https://bifrost.marvel.com/v1/catalog/comics/mu?byId={series_id}&byZone=marvel_site_zone&byType=comic_series&orderBy=release_date+desc&formatType=issue,digitalcomic,collection,digitalverticalcomic&limit=10000&offset=0&variants=false"
        )
        issue_ids = [issue["digital_id"] for issue in response.json()["data"]["results"]]
        return issue_ids


    async def _download_series_metadata(self, series_id: str) -> dict:
        """
        Download series metadata

        :param series_id: Id of comic series on marvel.com
        :returns: Dictionary with metadata
        """
        response = await self._client.get(
            f"https://gateway.marvel.com:443/v1/public/series/{series_id}?apikey={API_KEY}",
            headers = {
                "Referer": "https://developer.marvel.com/"
            }
        )
        return response.json()

    async def _get_issue_id(self, url: str) -> str:
        """
        Download issue id from url

        :param url: Url to issue info page
        :return: Issue id
        """
        response = await self._client.get(url)
        search = re.search(r"\"digitalComicID\":(\d+)", response.text)
        if not search:
            raise DataNotFound
        return search.group(1)



    async def download_book_from_id(self, issue_id: str) -> Book:
        return Book(
            metadata = await self._download_issue_metadata(issue_id),
            data = await self._download_issue_pages(issue_id)
        )


    async def _download_issue_metadata(self, issue_id: str) -> Metadata:
        """
        Download and parse metadata for issue

        :param issue_id: Identifier for issue
        :returns: Issue metadata
        """
        response = await self._client.get(
            f"https://bifrost.marvel.com/v1/catalog/digital-comics/metadata/{issue_id}"
        )
        issue_meta = response.json()["data"]["results"][0]["issue_meta"]
        return Metadata(
            title = issue_meta["title"],
            series = issue_meta.get("series_title"),
            description = issue_meta.get("description"),
            publisher = "Marvel",
            release_date = date.fromisoformat(issue_meta.get("release_date_digital")),
            authors = [c["full_name"] for c in issue_meta["creators"]["extended_list"]] if "extended_list" in issue_meta["creators"] else []
        )


    async def _download_issue_pages(self, issue_id: str) -> ImageList:
        """
        Download list of page links for issue

        :param issue_id: Identifier for issue
        :returns: List of links to comic pages
        """
        response = await self._client.get(
            f"https://bifrost.marvel.com/v1/catalog/digital-comics/web/assets/{issue_id}"
        )
        images = []
        for page in response.json()["data"]["results"][0]["pages"]:
            images.append(
                OnlineFile(
                    url = page["assets"]["source"],
                    extension = "jpg"
                )
            )
        return ImageList(images)



from .source import Source
from grawlix.book import Book, HtmlFile, HtmlFiles, OnlineFile, Metadata

from bs4 import BeautifulSoup

class RoyalRoad(Source):
    name: str = "Royal Road"
    match = [
        r"https://www.royalroad.com/fiction/\d+/[^/]+"
    ]
    _authentication_methods: list[str] = []


    async def download(self, url: str) -> Book:
        response = await self._client.get(url)
        soup = BeautifulSoup(response.text, "lxml")
        chapters = []
        for chapter in soup.find_all("tr", class_="chapter-row"):
            chapters.append(
                HtmlFile(
                    title = chapter.find("a").text.strip(),
                    file = OnlineFile(
                        url = f"https://royalroad.com{chapter.get('data-url')}",
                        extension = "html"
                    ),
                    selector = { "class": "chapter-content" }
                )
            )
        return Book(
            data = HtmlFiles(
                cover = OnlineFile(
                    url = soup.find("div", class_="cover-art-container") \
                        .find("img") \
                        .get("src") \
                        .replace("full", "large"),
                    extension = "jpg"
                ),
                htmlfiles = chapters
            ),
            metadata = Metadata(
                title = soup.find("meta", attrs={"name":"twitter:title"}).get("content"),
                authors = [ soup.find("meta", attrs={"name":"twitter:creator"}).get("content") ]
            ),
            overwrite = True
        )



from grawlix.book import HtmlFiles, HtmlFile, OnlineFile, Book, SingleFile, Metadata, EpubInParts
from grawlix.exceptions import UnsupportedOutputFormat
from .output_format import OutputFormat, Update

import asyncio
from bs4 import BeautifulSoup
import os
from ebooklib import epub
from zipfile import ZipFile
import rich

class Epub(OutputFormat):
    extension = "epub"
    input_types = [SingleFile, HtmlFiles, EpubInParts]


    async def download(self, book: Book, location: str, update: Update) -> None:
        if isinstance(book.data, SingleFile):
            await self._download_single_file(book, location, update)
        elif isinstance(book.data, HtmlFiles):
            await self._download_html_files(book.data, book.metadata, location, update)
        elif isinstance(book.data, EpubInParts):
            await self._download_epub_in_parts(book.data, book.metadata, location, update)
        else:
            raise UnsupportedOutputFormat


    async def _download_html_files(self, html: HtmlFiles, metadata: Metadata, location: str, update: Update) -> None:
        output = epub.EpubBook()
        output.set_title(metadata.title)
        for author in metadata.authors:
            output.add_author(author)
        file_count = len(html.htmlfiles) + 1 # Html files + cover

        async def download_cover(cover_file: OnlineFile):
            cover_filename = f"cover.{cover_file.extension}"
            epub_cover = epub.EpubCover(file_name = cover_filename)
            epub_cover.content = await self._download_file(cover_file)
            output.add_item(epub_cover)
            epub_cover_page = epub.EpubCoverHtml(image_name = cover_filename)
            if update:
                update(1/file_count)
            return epub_cover_page


        async def download_file(index: int, file: HtmlFile):
            response = await self._client.get(
                file.file.url,
                headers = file.file.headers,
                cookies = file.file.cookies,
                follow_redirects=True
            )
            soup = BeautifulSoup(response.text, "lxml")
            selected_element = soup.find(attrs=file.selector)
            epub_file = epub.EpubHtml(
                title = file.title,
                file_name = f"part {index}.html",
                content = str(selected_element)
            )
            if update:
                update(1/file_count)
            return epub_file

        # Download files
        tasks = [
            download_file(index, file)
            for index, file in enumerate(html.htmlfiles)
        ]
        if html.cover:
            tasks.append(download_cover(html.cover))
        epub_files = await asyncio.gather(*tasks)

        # Add files to epub
        for epub_file in epub_files:
            output.add_item(epub_file)
            output.spine.append(epub_file)
            output.toc.append(epub_file)

        # Complete book
        output.add_item(epub.EpubNcx())
        output.add_item(epub.EpubNav())
        epub.write_epub(location, output)


    async def _download_epub_in_parts(self, data: EpubInParts, metadata: Metadata, location: str, update: Update) -> None:
        files = data.files
        file_count = len(files)
        progress = 1/(file_count)
        temporary_file_location = f"{location}.tmp"

        added_files: set[str] = set()
        def get_new_files(zipfile: ZipFile):
            """Returns files in zipfile not already added to file"""
            for filename in zipfile.namelist():
                if filename in added_files or filename.endswith(".opf") or filename.endswith(".ncx"):
                    continue
                yield filename

        output = epub.EpubBook()
        for file in files:
            await self._download_and_write_file(file, temporary_file_location)
            with ZipFile(temporary_file_location, "r") as zipfile:
                for filepath in get_new_files(zipfile):
                    content = zipfile.read(filepath)
                    if filepath.endswith("html"):
                        filename = os.path.basename(filepath)
                        is_in_toc = False
                        title = None
                        for key, value in data.files_in_toc.items():
                            toc_filename = key.split("#")[0]
                            if filename == toc_filename:
                                title = value
                                is_in_toc = True
                                break
                        epub_file = epub.EpubHtml(
                            title = title,
                            file_name = filepath,
                            content = content
                        )
                        output.add_item(epub_file)
                        output.spine.append(epub_file)
                        if is_in_toc:
                            output.toc.append(epub_file)
                    else:
                        epub_file = epub.EpubItem(
                            file_name = filepath,
                            content = content
                        )
                        output.add_item(epub_file)
                    added_files.add(filepath)
            if update:
                update(progress)
        os.remove(temporary_file_location)

        output.add_item(epub.EpubNcx())
        output.add_item(epub.EpubNav())
        epub.write_epub(location, output)
        exit()



from .book import Book, Series
from .config import load_config, Config, SourceConfig
from .exceptions import SourceNotAuthenticated, GrawlixError, AccessDenied
from .sources import load_source, Source
from .output import download_book
from . import  arguments, logging

from typing import Tuple, Optional
from rich.prompt import Prompt
from rich.progress import Progress
from functools import partial
import os
import asyncio
import traceback


def get_or_ask(attr: str, hidden: bool, source_config: Optional[SourceConfig], options) -> str:
    return getattr(options, attr, None) \
    or getattr(source_config, attr, None) \
    or Prompt.ask(attr.capitalize(), password=hidden)


def get_login(source: Source, config: Config, options) -> Tuple[str, str, Optional[str]]:
    """
    Get login credentials for source

    :param source: Source to authenticate
    :param config: Content of config file
    :param options: Command line options
    :returns: Login credentials
    """
    source_name = source.name.lower().replace(" ", "")
    source_config = config.sources.get(source_name)

    username = get_or_ask("username", False, source_config, options)
    password = get_or_ask("password", True, source_config, options)
    if "library" in source._login_credentials:
        library = get_or_ask("library", False, source_config, options)
    # if source_name in config.sources:
    #     username = config.sources[source_name].username or options.username
    #     password = config.sources[source_name].password or options.password
    #     library = config.sources[source_name].library or options.library
    # else:
    #     username = options.username
    #     password = options.password
    #     library = options.library

    return username, password, library


def get_urls(options) -> list[str]:
    """
    Retrieves all available urls from input arguments
    - From urls argument
    - From file argument

    :param options: Cli options
    :returns: All urls listed in arguments
    """
    urls = []
    if options.urls:
        urls.extend(options.urls)
    if options.file:
        with open(options.file, "r") as f:
            content = f.read()
            urls.extend(content.split("\n"))
    return urls


def get_cookie_file(options) -> Optional[str]:
    """
    Get path to cookie file

    :param options: Cli arguments
    :returns: Path to cookie file
    """
    if options.cookie_file is not None and os.path.exists(options.cookie_file):
        return options.cookie_file
    if os.path.exists("./cookies.txt"):
        return "./cookies.txt"
    return None


async def authenticate(source: Source, config: Config, options):
    """
    Authenticate with source

    :param source: Source to authenticate
    :param config: Content of config file
    :param options: Command line options
    """
    logging.info(f"Authenticating with source [magenta]{source.name}[/]")
    if source.supports_login:
        username, password, library = get_login(source, config, options)
        await source.login(username, password, library=library)
        source.authenticated = True
    if not source.authenticated and source.supports_cookies:
        cookie_file = get_cookie_file(options)
        if cookie_file:
            source.load_cookies(cookie_file)
            source.authenticated = True
    if not source.authenticated:
        raise SourceNotAuthenticated


async def main() -> None:
    args = arguments.parse_arguments()
    config = load_config()
    logging.debug_mode = args.debug
    urls = get_urls(args)
    for url in urls:
        try:
            source: Source = load_source(url)
            if not source.authenticated and source.requires_authentication:
                await authenticate(source, config, args)
            result = await source.download(url)
            if isinstance(result, Book):
                with logging.progress(result.metadata.title, source.name) as progress:
                    template: str = args.output or "{title}.{ext}"
                    await download_with_progress(result, progress, template)
            elif isinstance(result, Series):
                await download_series(source, result, args)
            logging.info("")
        except GrawlixError as error:
            error.print_error()
            if logging.debug_mode:
                traceback.print_exc()
            exit(1)


async def download_series(source: Source, series: Series, args) -> None:
    """
    Download books in series

    :param series: Series to download
    """
    template = args.output or "{series}/{title}.{ext}"
    with logging.progress(series.title, source.name, len(series.book_ids)) as progress:
        for book_id in series.book_ids:
            try:
                book: Book = await source.download_book_from_id(book_id)
                await download_with_progress(book, progress, template)
            except AccessDenied as error:
                logging.info("Skipping - Access Denied")



async def download_with_progress(book: Book, progress: Progress, template: str):
    """
    Download book with progress bar in cli

    :param book: Book to download
    :param progress: Progress object
    :param template: Output template
    """
    task = logging.add_book(progress, book)
    update_function = partial(progress.advance, task)
    await download_book(book, update_function, template)
    progress.advance(task, 1)


def run() -> None:
    """Start main function"""
    asyncio.run(main())


if __name__ == "__main__":
    run()



from .output_format import OutputFormat, Update, Book
from grawlix.book import ImageList, OnlineFile
from grawlix.exceptions import UnsupportedOutputFormat
from .metadata.comicinfo import to_comic_info

from zipfile import ZipFile
import asyncio
import math

class Cbz(OutputFormat):
    """Comic book zip file"""

    extension: str = "cbz"
    input_types = [ImageList]

    async def download(self, book: Book, location: str, update: Update) -> None:
        if not isinstance(book.data, ImageList):
            raise UnsupportedOutputFormat
        semaphore = asyncio.Semaphore(10)
        images = book.data.images
        image_count = len(images)
        with ZipFile(location, mode="w") as zip:
            async def download_page(index: int, file: OnlineFile):
                async with semaphore:
                    content = await self._download_file(file)
                    padded_index = str(index).zfill(math.ceil(math.log10(image_count)))
                    zip.writestr(f"Image {padded_index}.{file.extension}", content)
                    if update:
                        update(1/image_count)
            tasks = [
                asyncio.create_task(download_page(index, file))
                for index, file in enumerate(images)
            ]
            await asyncio.wait(tasks)
            zip.writestr("ComicInfo.xml", to_comic_info(book.metadata))



from grawlix.book import Book, Metadata, SingleFile, OnlineFile, Result, Series
from grawlix.exceptions import SourceNotAuthenticated, InvalidUrl, DataNotFound
from grawlix import logging
from .source import Source

import json
import re
from urllib3.util import parse_url
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad
from typing import Any

class Storytel(Source):
    name: str = "Storytel"
    match = [
        r"https?://(?:www.)?(?:storytel|mofibo).com/(?P<language>\w+)(?:/(?P<language2>\w+))?/(?P<list_type>(?:books|series|authors|narrators|publishers|categories))/.+",
    ]
    _authentication_methods = [ "login" ]
    __download_counter = 0

    async def download(self, url: str) -> Result:
        await self.reauthenticate()

        if m := re.match(self.match[0], url):
            language, language2, list_type = m.groups()

            if list_type == "books":
                book_id = self.extract_id_from_url(url)
                logging.debug(f"{book_id=}")
                return await self.download_book_from_id(book_id)

            if list_type in ("series", "authors", "narrators"):
                return await self.download_list(url, list_type, language)

        raise InvalidUrl


    async def download_book_from_id(self, book_id: str) -> Book:
        # Epub location
        response = await self._client.get(
            f"https://api.storytel.net/assets/v2/consumables/{book_id}/ebook",
        )
        self.__download_counter += 1
        epub_url = response.headers["Location"]

        # Book details
        response = await self._client.get(
            f"https://api.storytel.net/book-details/consumables/{book_id}?kidsMode=false&configVariant=default"
        )
        details = response.json()

        return Book(
            metadata = Metadata(
                title = details["title"]
            ),
            data = SingleFile(
                OnlineFile(
                    url = epub_url,
                    extension = "epub",
                    headers = self._client.headers
                )
            )
        )


    async def download_list(self, url: str, list_type: str, language: str) -> Series:
        """
        Download list of books

        :param url: Url of list page
        :param list_type: Type of list (either series, authors, or narrators)
        :param language: The language of the books in the list
        :return: List of books
        """
        list_id = self.extract_id_from_url(url)
        logging.debug(f"{list_id=}")
        list_details = await self.download_list_details(list_id, list_type, language)

        books: list[str] = [
            item["id"]
            for item in list_details["items"]
            if "ebook" in [format["type"] for format in item["formats"]]
        ]

        return Series(
            title = list_details["title"],
            book_ids = books
        )


    async def download_list_details(
        self,
        list_id: str,
        list_type: str,
        languages: str,
        formats: str = "ebook",
    ) -> dict[str, Any]:
        """Download details about book list

        :param formats: comma serapted list of formats (abook,ebook,podcast)
        :param languages: comma seperated list of languages (en,de,tr,ar,ru,pl,it,es,sv,fr,nl)
        """
        nextPageToken = 0

        # API returns only 10 items per request
        # if the nextPageToken
        result: dict[str, Any] = {"nextPageToken": False}

        while result["nextPageToken"] is not None:
            params: dict[str, str] = {
                "includeListDetails": "true",  # include listMetadata,filterOptions,sortOption sections
                "includeFormats": formats,
                "includeLanguages": languages,
                "kidsMode": "false",
            }
            if result["nextPageToken"]:
                params["nextPageToken"] = result["nextPageToken"]

            response = await self._client.get(
                f"https://api.storytel.net/explore/lists/{list_type}/{list_id}",
                params=params,
            )

            data = response.json()
            if result["nextPageToken"] == 0:
                result = data
            else:
                result["items"].extend(data["items"])
                result["nextPageToken"] = data["nextPageToken"]
            logging.debug(f"{result=}")

        return result


    @staticmethod
    def extract_id_from_url(url: str) -> str:
        """
        Extract id from url

        :param url: Url containing id
        :return: Id
        """
        parsed = parse_url(url)
        if parsed.path is None:
            raise DataNotFound
        return parsed.path.split("-")[-1]


    @staticmethod
    def encrypt_password(password: str) -> str:
        """
        Encrypt password with predefined keys.
        This encrypted password is used for login.

        :param password: User defined password
        :returns: Encrypted password
        """
        # Thanks to https://github.com/javsanpar/storytel-tui
        key = b"VQZBJ6TD8M9WBUWT"
        iv = b"joiwef08u23j341a"
        msg = pad(password.encode(), AES.block_size)
        cipher = AES.new(key, AES.MODE_CBC, iv)
        cipher_text = cipher.encrypt(msg)
        return cipher_text.hex()


    async def reauthenticate(self) -> None:
        """Reauthenticate if required"""
        if self.__download_counter > 0 and self.__download_counter % 10 == 0:
            await self.authenticate()


    async def authenticate(self) -> None:
        """Authenticate with storytel"""
        response = await self._client.post(
            f"https://www.storytel.com/api/login.action?m=1&token=guestsv&userid=-1&version=23.49&terminal=android&locale=sv&deviceId=995f2562-0e44-4410-b1b9-8d08261f33c4&kidsMode=false",
            data = {
                "uid": self.__username,
                "pwd": self.__password
            }
        )
        if response.status_code != 200:
            raise SourceNotAuthenticated
        user_data = response.json()
        jwt = user_data["accountInfo"]["jwt"]
        self._client.headers.update({"authorization": f"Bearer {jwt}"})


    async def login(self, username: str, password: str, **kwargs) -> None:
        self.__username = username
        self.__password = self.encrypt_password(password)
        self._client.headers.update({"User-Agent": "Storytel/23.49 (Android 13; Pixel 6) Release/2288481"})
        await self.authenticate()



from grawlix.logging import print_error_file

ISSUE_URL = "https://github.com/jo1gi/grawlix/issues"
REPO_URL = "https://github.com/jo1gi/grawlix"

class GrawlixError(Exception):
    error_file: str

    def print_error(self) -> None:
        print_error_file(
            self.error_file,
            repo = REPO_URL,
            issue = ISSUE_URL,
        )


class DataNotFound(GrawlixError):
    error_file = "data_not_found"

class InvalidUrl(GrawlixError):
    error_file = "invalid_url"

class UnsupportedOutputFormat(GrawlixError):
    error_file = "unsupported_output_format"

class SourceNotAuthenticated(GrawlixError):
    error_file = "source_not_authenticated"

class ThrottleError(GrawlixError):
    error_file = "throttle"

class AccessDenied(GrawlixError):
    error_file = "access_denied"



from Crypto.Cipher import AES
from typing import Union, Protocol
from dataclasses import dataclass


@dataclass(slots=True)
class AESEncryption:
    key: bytes
    iv: bytes

    def decrypt(self, data: bytes) -> bytes:
        cipher = AES.new(self.key, AES.MODE_CBC, self.iv)
        return cipher.decrypt(data)


@dataclass(slots=True)
class AESCTREncryption:
    key: bytes
    nonce: bytes
    initial_value: bytes

    def decrypt(self, data: bytes) -> bytes:
        cipher = AES.new(
            key = self.key,
            mode = AES.MODE_CTR,
            nonce = self.nonce,
            initial_value = self.initial_value
        )
        return cipher.decrypt(data)


@dataclass(slots=True)
class XOrEncryption:
    key: bytes

    def decrypt(self, data: bytes) -> bytes:
        key_length = len(self.key)
        decoded = []
        for i in range(0, len(data)):
            decoded.append(data[i] ^ self.key[i % key_length])
        return bytes(decoded)


class Encryption(Protocol):
    def decrypt(self, data: bytes) -> bytes: ...


def decrypt(data: bytes, encryption: Encryption) -> bytes:
    """
    Decrypt data with specified encryption algorithm

    :param data: Bytes to decrypt
    :param encryption: Information about how to decrypt
    :returns: Decrypted data
    """
    return encryption.decrypt(data)



from .source import Source
from grawlix.book import Book, Metadata, ImageList, OnlineFile, Series, Result
from grawlix.exceptions import InvalidUrl, DataNotFound
from grawlix.logging import debug
from grawlix.utils import get_arg_from_url

import re
from urllib.parse import urlparse
from typing import Tuple, Optional

BASEURL = "https://reader.flipp.dk/html5/reader"

LANGUAGE_CODE_MAPPING = {
    "dk": "da-DK",
    "no": "nb-NO",
    "se": "sv-SE"
}

class Flipp(Source):
    name: str = "Flipp"
    match = [
        r"https?://reader.flipp.(dk|no|se)/html5/reader/production/default.aspx\?pubname=&edid=([^/]+)",
        r"https?://(magasiner|blader).flipp.(dk|no|se)/flipp/web-app/#/publications/.+"
    ]
    _authentication_methods: list[str] = []
    _login_cache: dict = {}



    async def download(self, url: str) -> Result:
        domain_extension = self.get_domain_extension(url)
        if re.match(self.match[0], url):
            issue_id = self._extract_issue_id(url)
            series_id = await self._get_series_id(issue_id)
            debug(f"{series_id=}")
            return await self._download_book(issue_id, series_id, domain_extension)
        elif re.match(self.match[1], url):
            return await self._download_series(url, domain_extension)
        raise InvalidUrl


    async def download_book_from_id(self, book_id: Tuple[str, str, str]) -> Book:
        series_id, issue_id, language_code = book_id
        return await self._download_book(issue_id, series_id, language_code)


    async def _download_series(self, url: str, language_code) -> Series:
        """
        Download series with book ids from Flipp

        :param url: Url of series
        :returns: Series object
        """
        series_id = url.split("/")[-1]
        login_info = await self._download_login_info(language_code)
        series_metadata = self._extract_series_data(login_info, series_id)
        issues = []
        for issue in series_metadata["issues"]:
            issue_id = issue["customIssueCode"]
            issues.append((series_id, issue_id, language_code))
        return Series(
            title = series_metadata["name"],
            book_ids = issues
        )


    async def _download_login_info(self, language_code: str) -> dict:
        """
        Download login info from Flipp
        Will use cache if available

        :returns: Login info
        """
        if language_code in self._login_cache:
            return self._login_cache[language_code]
        login_cache = await self._client.post(
            "https://flippapi.egmontservice.com/api/signin",
            headers = {
                "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:111.0) Gecko/20100101 Firefox/111.0"
            },
            json = {
                "email": "",
                "password": "",
                "token": "",
                "languageCulture": LANGUAGE_CODE_MAPPING[language_code],
                "appId": "",
                "appVersion": "",
                "uuid": "",
                "os": ""
            }
        )
        self._login_cache[language_code] = login_cache.json()
        return login_cache.json()


    def _extract_series_data(self, response: dict, series_id: str) -> dict:
        """
        Extract metadata about series from login response

        :param response: Login response from Flipp
        :param series_id: Id of series
        :returns: Metadata about series
        """
        for publication in response["publications"]:
            if publication["customPublicationCode"] == series_id:
                return publication
        raise DataNotFound


    async def _download_book(self, issue_id: str, series_id: str, language_code: str) -> Book:
        """
        Download book from Flipp

        :param issue_id: Issue identifier
        :param series_id: Series identifier
        :returns: Book metadata
        """
        pages = await self._get_pages(issue_id, series_id)
        metadata = await self._get_metadata(issue_id, series_id, language_code)
        return Book(
            data = ImageList(pages),
            metadata = Metadata(
                title = f"{metadata['series_name']} {metadata['issueName']}",
                series = metadata["series_name"],
                identifier = issue_id
            ),
        )


    async def _get_metadata(self, issue_id: str, series_id: str, language_code: str) -> dict:
        """
        Download and extract issue data

        :param issue_id: Issue id
        :param series_id: Series id
        :returns: Issue metadata
        """
        login_info = await self._download_login_info(language_code)
        series_metadata = self._extract_series_data(login_info, series_id)
        for issue in series_metadata["issues"]:
            if issue["customIssueCode"] == issue_id:
                issue["series_name"] = series_metadata["name"]
                return issue
        raise DataNotFound


    @staticmethod
    def get_domain_extension(url: str) -> str:
        """
        Extract domain extension from url

        :param url: Url to parse
        :returns: Domain extension of url
        """
        parsed_url = urlparse(url)
        extension = parsed_url.netloc.split(".")[-1]
        return extension


    @staticmethod
    def _extract_issue_id(url: str) -> str:
        """
        Extract eid from url

        :param url: Url to extract data from
        :returns: Eid in url
        """
        return get_arg_from_url(url, "edid")


    async def _get_series_id(self, issue_id: str) -> str:
        """
        Download series id from issue id

        :param issue_id: Issue id
        :returns: Series id
        """
        response = await self._client.get(f"{BASEURL}/production/default.aspx?pubname=&edid={issue_id}")
        # TODO Make faster
        search = re.search(r'publicationguid = "([^"]+)', response.text)
        if search is None:
            raise DataNotFound
        return search.group(1)


    async def _get_pages(self, issue_id: str, series_id: str) -> list[OnlineFile]:
        """
        Download page metadata for book

        :param issue_id: Issue id
        :param series_id: Series id
        :return: Page image links
        """
        response = await self._client.get(
            f"{BASEURL}/get_page_groups_from_eid.aspx?pubid={series_id}&eid={issue_id}",
        )
        result = []
        for page in response.json()["pageGroups"]:
            # Find image id in low quality image url
            low_quality_url = urlparse(page["pages"][0]["image"])
            image_id = low_quality_url.path[1:-9]
            high_quality_url = f"http://pages.cdn.pagesuite.com/{image_id}/highpage.jpg?method=true"
            result.append(OnlineFile(high_quality_url, "jpg"))
        return result



from grawlix.book import Book, Metadata, SingleFile, OnlineFile
from grawlix import AESEncryption
from grawlix.exceptions import ThrottleError

import re
from .source import Source

class Saxo(Source):
    name: str = "Saxo"
    match = [
        r"https://(\w+\.)?saxo.(com|dk)/.+\d+$"
    ]
    _authentication_methods = [ "login" ]
    user_id: str

    async def login(self, username: str, password: str, **kwargs) -> None:
        response = await self._client.post(
            "https://auth-read.saxo.com/auth/token",
            data = {
                "username": username,
                "password": password,
                "grant_type": "password",
            },
            headers = {
                "Content-Type": "application/x-www-form-urlencoded"
            }
        )
        json = response.json()
        bearer_token = json["access_token"]
        self._client.headers = {
            "Appauthorization": f"bearer {bearer_token}",
            "App-Os": "android",
            "App-Version": "6.2.4"
        }
        self.user_id = json["id"]


    async def download(self, url: str) -> Book:
        isbn = self._extract_isbn_from_url(url)
        book_id = await self._get_book_id(isbn)
        metadata = await self._get_book_metadata(book_id)
        ebook_id = metadata["id"] # Id of ebook file
        return Book(
            metadata = self._extract_metadata(metadata),
            data = SingleFile(
                OnlineFile(
                    url = await self._get_book_file_link(ebook_id),
                    extension = "epub",
                    # Encryption keys extracted from app
                    encryption = AESEncryption(
                        key = b"CD3E9D141D8EFC0886912E7A8F3652C4",
                        iv = b"78CB354D377772F1"
                    )
                )
            )
        )


    async def _get_book_id(self, isbn: str) -> str:
        """
        Download internal book id of book from isbn

        :param isbn: Isbn of book
        :returns: Saxo internal book id
        """
        response = await self._client.get(
            f"https://api-read.saxo.com/api/v2/search/user/{self.user_id}/premium/books/{isbn}"
        )
        return response.json()["items"][0]["bookId"]


    async def _get_book_metadata(self, book_id: str) -> dict:
        """
        Download metadata of book

        :param book_id: Id of book
        :returns: Metadata of book
        """
        response = await self._client.get(
            f"https://api-read.saxo.com/api/v2/book/{book_id}/user/{self.user_id}/details"
        )
        return response.json()["ebooks"][0]


    async def _get_book_file_link(self, ebook_id: str) -> str:
        """
        Download link to epub file

        :param ebook_id: Id of ebook file
        :returns: Link to ebook file
        :raises ThrottleError: If there have been too many downloads
        """
        response = await self._client.get(
            f"https://api-read.saxo.com/api/v1/book/{ebook_id}/content/encryptedstream/"
        )
        json = response.json()
        if not "link" in json:
            raise ThrottleError
        return json["link"]


    @staticmethod
    def _extract_metadata(metadata: dict) -> Metadata:
        """
        Extract metadata from metadata response from Saxo

        :param metadata: Metadata response from saxo
        :returns: Metadata formatted as `grawlix.Metadata`
        """
        return Metadata(
            title = metadata["title"],
            authors = [metadata["author"]] if "author" in metadata else [],
            language = metadata.get("languageLocalized")
        )


    @staticmethod
    def _extract_isbn_from_url(url: str) -> str:
        """
        Extracts isbn from url

        :param url: Url of book
        :returns: Isbn of book
        """
        isbn_match = re.search(f"\d+$", url)
        if isbn_match and isbn_match.group():
            return isbn_match.group()
        raise NotImplemented



from grawlix import __version__

import argparse

def parse_arguments() -> argparse.Namespace:
    # Help
    parser = argparse.ArgumentParser(
        prog = "grawlix",
        description = "Download ebooks"
    )
    parser.add_argument(
        '-v',
        '--version',
        action = "version",
        version = f"grawlix {__version__}"
    )
    # Basics
    parser.add_argument(
        'urls',
        help = "Links to ebooks",
        nargs = "*"
    )
    parser.add_argument(
        '-f',
        '--file',
        help = "File with links (One link per line)",
        dest = "file"
    )
    # Authentication
    parser.add_argument(
        '-u',
        '--username',
        help = "Username for login",
        dest = "username",
    )
    parser.add_argument(
        '-p',
        '--password',
        help = "Password for login",
        dest = "password",
    )
    parser.add_argument(
        '--library',
        help = "Library for login",
        dest = "library",
    )
    parser.add_argument(
        '-c',
        '--cookies',
        help = "Path to netscape cookie file",
        dest = "cookie_file"
    )
    # Outputs
    parser.add_argument(
        '-o',
        '--output',
        help = "Output destination",
        dest = "output"
    )
    # Logging
    parser.add_argument(
        '--debug',
        help = "Enable debug messages",
        dest = "debug",
        action="store_true",
    )
    return parser.parse_args()



from grawlix.book import Book, SingleFile
from .output_format import OutputFormat, Update
import shutil
import subprocess

class Acsm(OutputFormat):
    extension = "acsm"
    input_types = [SingleFile]

    async def download(self, book: Book, location: str, update_func: Update) -> None:
        # Download and write acsm file to disk
        await self._download_single_file(book, location, update_func)
        # TODO: Implement more general solution
        # Decrypt if knock is available
        # https://web.archive.org/web/20221016154220/https://github.com/BentonEdmondson/knock
        if shutil.which("knock") is not None:
            subprocess.run(
                ["knock", location],
                capture_output = True
            )
        else:
            # TODO: Print warning
            pass



from grawlix.exceptions import DataNotFound

from urllib.parse import urlparse, parse_qs
from functools import lru_cache
import importlib.resources

def get_arg_from_url(url: str, key: str) -> str:
    parsed_url = urlparse(url)
    query = parse_qs(parsed_url.query)
    try:
        return query[key][0]
    except:
        raise DataNotFound


@lru_cache
def levenstein_distance(a: str, b: str) -> int:
    """
    Calculates the levenstein distance between `a` and `b`

    https://en.wikipedia.org/wiki/Levenshtein_distance
    """
    if len(a) == 0:
        return len(b)
    if len(b) == 0:
        return len(a)
    if a[0] == b[0]:
        return levenstein_distance(a[1:], b[1:])
    return 1 + min(
        levenstein_distance(a, b[1:]), # Character is inserted
        levenstein_distance(a[1:], b), # Character is deleted
        levenstein_distance(a[1:], b[1:]) # Character is replaced
    )



def nearest_string(input: str, list: list[str]) -> str:
    """
    Finds the nearest string in `list` to `input` based on levenstein distance
    """
    return sorted(list, key = lambda x: levenstein_distance(input, x))[0]


def read_asset_file(path: str) -> str:
    """
    Read asset file from the grawlix module

    :param path: Path relative to root of grawlix module
    """
    return importlib.resources.files("grawlix") \
        .joinpath(path) \
        .read_text(encoding="utf8")



from .source import Source
from grawlix.book import Book, Metadata, ImageList, OnlineFile, Series, Result
from grawlix.utils import get_arg_from_url
from grawlix.exceptions import InvalidUrl

from bs4 import BeautifulSoup
import re
from urllib.parse import urlparse

USER_AGENT = "Mozilla/5.0 (X11; Linux x86_64; rv:111.0) Gecko/20100101 Firefox/111.0"
MOBILE_USER_AGENT = "Mozilla/5.0 (iPhone; CPU iPhone OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Mobile/15E148 Safari/604.1"

class Webtoons(Source[str]):

    name: str = "Webtoons"
    match = [
        r"https://www.webtoons.com/../.+/.+/.+/viewer\?title_no=\d+&episode_no=\d+",
        r"https://www.webtoons.com/../.+/.+/list\?title_no=\d+"
    ]
    _authentication_methods: list[str] = []

    async def download(self, url: str) -> Result[str]:
        if re.match(self.match[0], url):
            return await self._download_episode(url)
        if re.match(self.match[1], url):
            return await self._download_series(url)
        raise InvalidUrl


    async def download_book_from_id(self, book_id: str) -> Book:
        return await self._download_episode(book_id)


    async def _download_series(self, url: str) -> Series[str]:
        """
        Download a series of webtoons

        :param url: Url of series
        :returns: Webtoons series data
        """
        parsed_url = urlparse(url)
        response = await self._client.get(
            f"https://m.webtoons.com{parsed_url.path}",
            params = parsed_url.query,
            headers = {
                "User-Agent": MOBILE_USER_AGENT,
            },
            cookies = {
                "needGDPR": "FALSE",
                "needCCPA": "FALSE",
                "needCOPPA": "FALSE"
            },
            follow_redirects = True,
        )
        soup = BeautifulSoup(response.text, "lxml")
        title = soup.find("meta", property="og:title").get("content")
        episodes = []
        for episode in soup.find_all("li", class_="_episodeItem"):
            episode_link = episode.find("a").get("href")
            episodes.append(episode_link)
        return Series(
            title,
            book_ids = episodes
        )


    async def _download_episode(self, url: str) -> Book:
        """
        Download single webtoon episode

        :param url: Url of episode
        :returns: Episode
        """
        response = await self._client.get(url, follow_redirects = True)
        soup = BeautifulSoup(response.text, "lxml")
        title = soup.find("h1", class_="subj_episode").get("title")
        series = soup.find("div", class_="subj_info").find("a").get("title")
        images = []
        for image in soup.find("div", class_="viewer_img _img_viewer_area").find_all("img"):
            images.append(
                OnlineFile(
                    url = image.get("data-url"),
                    extension = "png",
                    headers = { "Referer": "https://www.webtoons.com/" }
                )
            )
        return Book(
            data = ImageList(images),
            metadata = Metadata(
                title,
                series = series
            )
        )



from grawlix.book import Book

import rich
from rich.console import Console
from rich.markup import render
from rich.progress import Progress, BarColumn, ProgressColumn, TaskID, SpinnerColumn
from rich.style import Style

from typing import Union
from dataclasses import dataclass
import importlib.resources

console = Console(stderr=True)

debug_mode = False
DEBUG_PREFIX = render("[yellow bold]DEBUG[/]")


def debug(msg: str, remove_styling=False) -> None:
    """
    Print debug message in console

    :param msg: Message to print
    :param remove_styling: Remove automated styling from message
    """
    if debug_mode:
        if remove_styling:
            rendered_msg = render(msg, style=Style(bold=False, color="white"))
            console.print(DEBUG_PREFIX, rendered_msg)
        else:
            console.print(DEBUG_PREFIX, msg)


def info(msg: str) -> None:
    """
    Print message in console

    :param msg: Message to print
    """
    console.print(msg)


def error(msg: str) -> None:
    console.print(msg)


def print_error_file(name: str, **kwargs) -> None:
    """
    Print predefined error message

    :param name: Name of error file
    """
    message = importlib.resources.files("grawlix") \
        .joinpath(f"assets/errors/{name}.txt") \
        .read_text("utf8") \
        .format(**kwargs) \
        .strip()
    error(message)


def progress(category_name: str, source_name: str, count=1) -> Progress:
    if count > 1:
        console.print(f"Downloading [yellow not bold]{count}[/] books in [blue]{category_name}[/] from [magenta]{source_name}[/]")
    else:
        console.print(f"Downloading [blue bold]{category_name}[/] from [magenta]{source_name}[/]")
    progress = Progress(
        SpinnerColumn(),
        "{task.description}",
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        console = console,
    )
    return progress


def add_book(progress: Progress, book: Book) -> TaskID:
    task = progress.add_task(
        f"[blue]{book.metadata.title}[/]",
        total = 1
    )
    return task



from grawlix.book import Result, Book, SingleFile, Metadata, OnlineFile
from grawlix.encryption import AESCTREncryption
from grawlix.exceptions import InvalidUrl, DataNotFound
from grawlix.utils import nearest_string
from .source import Source

from bs4 import BeautifulSoup
import json
import re
from Crypto.Cipher import AES
from base64 import b64decode

LOGIN_PAGE_URL = "https://ereolen.dk/adgangsplatformen/login?destination=/user"
KEY_ENCRYPTION_KEY = bytes([30, 193, 150, 69, 32, 247, 35, 95, 92, 255, 193, 159, 121, 40, 151, 179, 39, 159, 75, 110, 32, 205, 210, 58, 81, 55, 158, 33, 8, 149, 108, 74])
KEY_ENCRYPTION_IV = bytes([0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])

class Ereolen(Source):
    name: str = "eReolen"
    match: list[str] = [
        r"https://ereolen.dk/ting/object/\d+-.+/read/?$",
        r"https://ereolen.dk/ting/object/\d+-[^/]+/?$"
    ]
    _authentication_methods = [ "login" ]
    _login_credentials = [ "username", "password", "library" ]


    async def login(self, username: str, password: str, **kwargs) -> None:
        library = kwargs["library"]
        login_page = await self._client.get(LOGIN_PAGE_URL, follow_redirects=True)
        login_soup = BeautifulSoup(login_page.text, "lxml")
        borchk_login_form = login_soup.find(id="borchk-login-form")
        login_path = borchk_login_form.get("action")
        library_attr_name = borchk_login_form.find("label").get("for")
        libraries = self._extract_available_libraries(login_page.text)
        if not library in libraries:
            library = nearest_string(library, list(libraries.keys()))
        await self._client.post(
            f"https://login.bib.dk{login_path}",
            headers = { "Content-Type": "application/x-www-form-urlencoded" },
            data = {
                library_attr_name: library,
                "agency": libraries[library],
                "loginBibDkUserId": username,
                "pincode": password
            },
            follow_redirects = True
        )


    def _extract_available_libraries(self, login_page: str) -> dict[str, str]:
        """
        Extract list of available libraries from login page

        :param login_page: Content of login page as string
        :returns: Dictionary with name and id of each library
        """
        match = re.search("libraries = ({.+})<", login_page)
        if match is None:
            raise DataNotFound
        library_data = json.loads(match.group(1))
        libraries: dict[str, str] = {}
        for library in library_data["folk"]:
            library_name = library["name"]
            library_id = library["branchId"]
            libraries[library_name] = library_id
        return libraries


    async def download(self, url: str) -> Result:
        book_id: str = await self._get_book_id(url)
        metadata_response = await self._client.get(
            f"https://bookstreaming.pubhub.dk/v1/order/metadata/{book_id}",
        )
        metadata = metadata_response.json()
        key = self._decrypt_key(metadata["key"])
        return Book(
            data = SingleFile(
                OnlineFile(
                    url = f"https://bookstreaming.pubhub.dk/v1/order/file/{book_id}",
                    extension = "epub",
                    encryption = AESCTREncryption(
                        key,
                        nonce = bytes([0,0,0,0,0,0,0,0]),
                        initial_value = bytes([0,0,0,0,0,0,0,0])
                    )
                )
            ),
            metadata = Metadata(
                title = metadata["title"],
                authors = [ metadata["author"] ]
            )
        )


    def _decrypt_key(self, key: str) -> bytes:
        """
        Decrypt book encryption key

        :param key: Base64 encoded and encrypted key
        :returns: Decoded and decrypted key
        """
        decoded_key = b64decode(key)
        cipher = AES.new(KEY_ENCRYPTION_KEY, AES.MODE_CBC, KEY_ENCRYPTION_IV)
        return cipher.decrypt(decoded_key)[:16]


    async def _get_book_id(self, url: str) -> str:
        """
        Download and extract book_id

        :param url: Url to book page
        :returns: Book id
        """
        if re.match(self.match[0], url):
            if url.endswith("/"):
                url = url[:-1]
            return await self._get_book_id_from_reader(url)
        if re.match(self.match[1], url):
            return await self._get_book_id_from_reader(f"{url}/read")
        else:
            raise InvalidUrl


    async def _get_book_id_from_reader(self, url: str) -> str:
        """
        Download and extract book_id from reader page

        :param url: Url to reader page
        :returns: Book id
        """
        page = await self._client.get(url)
        soup = BeautifulSoup(page.text, "lxml")
        return soup.find("div", id="pubhub-reader").get("order-id")



from grawlix.book import Book, SingleFile, OnlineFile, ImageList, HtmlFiles, Book, OfflineFile, BookData
from grawlix.exceptions import UnsupportedOutputFormat
from grawlix.encryption import decrypt

import httpx
from typing import Callable, Optional

Update = Optional[Callable[[float], None]]

class OutputFormat:
    # Extension for output files
    extension: str
    input_types: list[type[BookData]]

    def __init__(self) -> None:
        self._client = httpx.AsyncClient()


    async def close(self) -> None:
        """Cleanup"""
        await self._client.aclose()


    async def download(self, book: Book, location: str, update_func: Update) -> None:
        """
        Download book

        :param book: Book to download
        :param location: Path to where the file is written
        :param update_func: Function to update progress bar
        """
        raise UnsupportedOutputFormat


    async def _download_single_file(self, book: Book, location: str, update_func: Update) -> None:
        """
        Download and write an `grawlix.SingleFile` to disk

        :param book: Book to download
        :param location: Path to where the file is written
        :raises UnsupportedOutputFormat: If datatype is not supported by format
        """
        if not isinstance(book.data, SingleFile):
            raise UnsupportedOutputFormat
        if not book.data.file.extension == self.extension:
            raise UnsupportedOutputFormat
        if isinstance(book.data.file, OnlineFile):
            await self._download_and_write_file(book.data.file, location, update_func)
        elif isinstance(book.data.file, OfflineFile):
            self._write_offline_file(book.data.file, location)


    async def _download_file(self, file: OnlineFile, update: Update = None) -> bytes:
        """
        Download `grawlix.OnlineFile` 

        :param file: File to download
        :param update: Update function that is called with a percentage every time a chunk is downloaded
        :returns: Content of downloaded file
        """
        content = b""
        async with self._client.stream("GET", file.url, headers = file.headers, cookies = file.cookies, follow_redirects=True) as request:
            total_filesize = int(request.headers["Content-length"])
            async for chunk in request.aiter_bytes():
                content += chunk
                if update:
                    update(len(chunk)/total_filesize)
            if file.encryption is not None:
                content = decrypt(content, file.encryption)
        return content


    async def _download_and_write_file(self, file: OnlineFile, location: str, update: Update = None) -> None:
        """
        Download `grawlix.OnlineFile` and write to content to disk

        :param file: File to download
        :param location: Path to where the file is written
        :param update: Update function that is called with a percentage every time a chunk is downloaded
        """
        content = await self._download_file(file, update)
        with open(location, "wb") as f:
            f.write(content)


    def _write_offline_file(self, file: OfflineFile, location: str) -> None:
        """
        Write the content of an `OfflineFile` to disk

        :param file: File to write to disk
        :param location: Path to where the file is written
        """
        with open(location, "wb") as f:
            content = file.content
            if file.encryption:
                content = decrypt(content, file.encryption)
            f.write(content)



from grawlix.exceptions import InvalidUrl

from .source import Source
from .dcuniverseinfinite import DcUniverseInfinite
from .ereolen import Ereolen
from .fanfictionnet import FanfictionNet
from .flipp import Flipp
from .internet_archive import InternetArchive
from .mangaplus import MangaPlus
from .marvel import Marvel
from .nextory import Nextory
from .royal_road import RoyalRoad
from .saxo import Saxo
from .storytel import Storytel
from .webtoons import Webtoons

import re

source_cache: dict[str, Source] = {}

def load_source(url: str) -> Source:
    """
    Find source that matches url
    Will save loaded sources in cache

    :param url: Url of book to download
    :returns: Source type for downloading url
    """
    source_cls = find_source(url)
    source_name = source_cls.name
    if source_name in source_cache:
        return source_cache[source_name]
    else:
        source = source_cls()
        source_cache[source_name] = source
        return source


def find_source(url: str) -> type[Source]:
    """
    Find source that matches url

    :param url: Url of book to download
    :returns: Source type for downloading url
    """
    for cls in get_source_classes():
        for num, match in enumerate(cls.match):
            if re.match(match, url):
                return cls
    raise InvalidUrl


def get_source_classes() -> list[type[Source]]:
    """
    Get all source types

    :returns: A list of all available source types
    """
    return [
        DcUniverseInfinite,
        Ereolen,
        FanfictionNet,
        Flipp,
        InternetArchive,
        MangaPlus,
        Marvel,
        Nextory,
        RoyalRoad,
        Saxo,
        Storytel,
        Webtoons
    ]



from .source import Source
from grawlix.encryption import XOrEncryption
from grawlix.book import Book, Metadata, ImageList, OnlineFile, Series, Result
from grawlix.exceptions import InvalidUrl

import re
import blackboxprotobuf
import json
import rich

class MangaPlus(Source):
    name: str = "Manga Plus"
    match = [
        r"https?://mangaplus.shueisha.co.jp/viewer/\d+",
        r"https?://mangaplus.shueisha.co.jp/titles/\d+"
    ]
    _authentication_methods: list[str] = []


    async def download(self, url: str) -> Result:
        if re.match(self.match[0], url):
            issue_id = url.split('/')[-1]
            return await self._download_issue(issue_id)
        if re.match(self.match[1], url):
            series_id = url.split("/")[-1]
            return await self._download_series(series_id)
        raise InvalidUrl


    async def download_book_from_id(self, book_id: str) -> Book:
        return await self._download_issue(book_id)


    async def _download_series(self, series_id: str) -> Series:
        """
        Download series from Manga Plus

        :param series_id: Identifier for series
        :returns: Series data
        """
        response = await self._client.get(
            f"https://jumpg-api.tokyo-cdn.com/api/title_detailV2",
            params = {
                "title_id": series_id,
                "lang": "eng",
                "os": "android",
                "os_ver": "32",
                "app_ver": "40",
                "secret": "2afb69fbb05f57a1856cf75e1c4b6ee6"
            },
        )
        data, _ = blackboxprotobuf.protobuf_to_json(response.content)
        parsed = json.loads(data)
        title = parsed["1"]["8"]["1"]["2"]
        issues = []
        def add_issues(data: dict, main: str):
            if main in data:
                x = data[main]
                if isinstance(x, list):
                    for i in x:
                        issues.append(i["2"])
                else:
                    issues.append(x["2"])
        for a in parsed["1"]["8"]["28"]:
            add_issues(a, "2")
            add_issues(a, "3")
            add_issues(a, "4")
        return Series(
            title,
            book_ids = issues
        )

    async def _download_issue(self, issue_id: str) -> Book:
        """
        Download issue from Manga Plus

        :param issue_id: Identifier for issue
        :returns: Issue metadata
        """
        url = f"https://jumpg-webapi.tokyo-cdn.com/api/manga_viewer?chapter_id={issue_id}&split=yes&img_quality=super_high"
        response = await self._client.get(url)
        content, _ = blackboxprotobuf.protobuf_to_json(response.content)
        images = []
        parsed = json.loads(content)
        for image in parsed["1"]["10"]["1"]:
            if "1" in image:
                images.append(
                    OnlineFile(
                        image["1"]["1"],
                        extension = "jpg",
                        encryption = XOrEncryption(bytes.fromhex(image["1"]["5"]))
                    )
                )
            elif "3" in image:
                title = image["3"]["1"]["4"]
        return Book(
            data = ImageList(images),
            metadata = Metadata(
                title,
                series = parsed["1"]["10"]["5"]
            )
        )



from grawlix import Encryption
from dataclasses import dataclass, field
from typing import Optional, Union, TypeVar, Generic, Any
from datetime import date

@dataclass(slots=True)
class Metadata:
    """Metadata about a book"""
    title: str
    series: Optional[str] = None
    index: Optional[int] = None
    authors: list[str] = field(default_factory=list)
    language: Optional[str] = None
    publisher: Optional[str] = None
    identifier: Optional[str] = None
    description: Optional[str] = None
    release_date: Optional[date] = None

    def as_dict(self) -> dict:
        return {
            "title": self.title,
            "series": self.series or "UNKNOWN",
            "index": self.index or "UNKNOWN",
            "publisher": self.publisher or "UNKNOWN",
            "identifier": self.identifier or "UNKNOWN",
            "language": self.language or "UNKNOWN",
            "authors": "; ".join(self.authors),
            "description": self.description or "UNKNOWN",
            "release_date": self.release_date.isoformat() if self.release_date else "UNKNOWN",
        }


@dataclass(slots=True)
class OnlineFile:
    """Instructions for downloading an online file"""
    url: str
    extension: str
    encryption: Optional[Encryption] = None
    headers: Optional[dict[str, str]] = None
    cookies: Optional[Any] = None # TODO Change type

@dataclass(slots=True)
class OfflineFile:
    """Stores content of a file"""
    content: bytes
    extension: str
    encryption: Optional[Encryption] = None

File = Union[OnlineFile, OfflineFile]


@dataclass(slots=True)
class SingleFile:
    """Bookdata in the form of a single file"""
    file: File


@dataclass(slots=True)
class ImageList:
    """
    List of images
    Mostly used for comic books
    """
    images: list[OnlineFile]


@dataclass(slots=True)
class EpubInParts:
    """
    Epub split up into smaller epubs
    """
    files: list[OnlineFile]
    files_in_toc: dict[str, str]


@dataclass(slots=True)
class HtmlFile:
    title: str
    file: OnlineFile
    selector: Optional[dict[str, str]]

@dataclass(slots=True)
class HtmlFiles:
    htmlfiles: list[HtmlFile]
    cover: Optional[OnlineFile] = None

BookData = Union[
    EpubInParts,
    SingleFile,
    ImageList,
    HtmlFiles
]

@dataclass(slots=True)
class Book:
    """Stores information about a book"""
    metadata: Metadata
    data: BookData
    overwrite: bool = False

T = TypeVar("T")

@dataclass(slots=True)
class Series(Generic[T]):
    """Stores a series of books"""
    title: str
    book_ids: list[T]

Result = Union[
    Book,
    Series[T]
]



from grawlix.book import Book, Metadata, OnlineFile, BookData, OnlineFile, SingleFile, EpubInParts, Result, Series
from grawlix.encryption import AESEncryption
from grawlix.exceptions import InvalidUrl
from .source import Source

from typing import Optional
import uuid
import rich
import base64

LOCALE = "en_GB"

class Nextory(Source):
    name: str = "Nextory"
    match = [
        r"https?://((www|catalog-\w\w).)?nextory.+"
    ]
    _authentication_methods = [ "login" ]


    @staticmethod
    def _create_device_id() -> str:
        """Create unique device id"""
        return str(uuid.uuid3(uuid.NAMESPACE_DNS, "audiobook-dl"))


    async def login(self, username: str, password: str, **kwargs) -> None:
        # Set permanent headers
        device_id = self._create_device_id()
        self._client.headers.update(
            {
                "X-Application-Id": "200",
                "X-App-Version": "5.0.0",
                "X-Locale": LOCALE,
                "X-Model": "Personal Computer",
                "X-Device-Id": device_id,
                "locale": LOCALE,
                "device": device_id,
                "osinfo": "Android 13",
                "model": "Personal Computer",
                "version": "4.34.6",
                "appid": "200",
            }
        )
        # Login for account
        session_response = await self._client.post(
            "https://api.nextory.com/user/v1/sessions",
            json = {
                "identifier": username,
                "password": password
            },
        )
        session_response = session_response.json()
        rich.print(session_response)
        login_token = session_response["login_token"]
        country = session_response["country"]
        self._client.headers.update(
            {
                "token": login_token,
                "X-Login-Token": login_token,
                "X-Country-Code": country,
            }
        )
        # Login for user
        profiles_response = await self._client.get(
            "https://api.nextory.com/user/v1/me/profiles",
        )
        profiles_response = profiles_response.json()
        rich.print(profiles_response)
        profile = profiles_response["profiles"][0]
        login_key = profile["login_key"]
        authorize_response = await self._client.post(
            "https://api.nextory.com/user/v1/profile/authorize",
            json = {
                "login_key": login_key
            }
        )
        authorize_response = authorize_response.json()
        rich.print(authorize_response)
        profile_token = authorize_response["profile_token"]
        self._client.headers.update({"X-Profile-Token": profile_token})
        self._client.headers.update({"X-Profile-Token": profile_token})


    @staticmethod
    def _find_epub_id(product_data) -> str:
        """Find id of book format of type epub for given book"""
        for format in product_data["formats"]:
            if format["type"] == "epub":
                return format["identifier"]
        raise InvalidUrl


    @staticmethod
    def _extract_id_from_url(url: str) -> str:
        """
        Extract id of book from url. This id is not always the internal id for
        the book.

        :param url: Url to book information page
        :return: Id in url
        """
        return url.split("-")[-1].replace("/", "")


    async def download(self, url: str) -> Result:
        url_id = self._extract_id_from_url(url)
        if "serier" in url:
            return await self._download_series(url_id)
        else:
            book_id = await self._get_book_id_from_url_id(url_id)
            return await self._download_book(book_id)


    async def download_book_from_id(self, book_id: str) -> Book:
        return await self._download_book(book_id)


    async def _download_series(self, series_id: str) -> Series:
        """
        Download series from Nextory

        :param series_id: Id of series on Nextory
        :returns: Series data
        """
        response = await self._client.get(
            f"https://api.nextory.com/discovery/v1/series/{series_id}/products",
            params = {
                "content_type": "book",
                "page": 0,
                "per": 100,
            }
        )
        series_data = response.json()
        book_ids = []
        for book in series_data["products"]:
            book_id = book["id"]
            book_ids.append(book_id)
        return Series(
            title = series_data["products"][0]["series"]["name"],
            book_ids = book_ids,
        )


    @staticmethod
    def _extract_series_name(product_info: dict) -> Optional[str]:
        if not "series" in product_info:
            return None
        return product_info["series"]["name"]


    async def _get_book_id_from_url_id(self, url_id: str) -> str:
        """
        Download book id from url id

        :param url_id: Id of book from url
        :return: Book id
        """
        response = await self._client.get(
            f"https://api.nextory.se/api/app/product/7.5/bookinfo",
            params = { "id": url_id },
        )
        rich.print(response.url)
        rich.print(response.content)
        exit()


    async def _download_book(self, book_id: str) -> Book:
        product_data = await self._client.get(
            f"https://api.nextory.com/library/v1/products/{book_id}"
        )
        product_data = product_data.json()
        epub_id = self._find_epub_id(product_data)
        pages = await self._get_pages(epub_id)
        return Book(
            data = pages,
            metadata = Metadata(
                title = product_data["title"],
                authors = [author["name"] for author in product_data["authors"]],
                series = self._extract_series_name(product_data),
            )
        )


    @staticmethod
    def _fix_key(value: str) -> bytes:
        """Remove unused data and decode key"""
        return base64.b64decode(value[:-1])


    async def _get_pages(self, epub_id: str) -> BookData:
        """
        Download page information for book

        :param epub_id: Id of epub file
        :return: Page data
        """
        # Nextory books are for some reason split up into multiple epub files -
        # one for each chapter file. All of these files has to be decrypted and
        # combined afterwards. Many of the provided epub files contain the same
        # files and some of them contain the same file names but with variation
        # in the content and comments that describe what should have been there
        # if the book was whole from the start.
        response = await self._client.get(
            f"https://api.nextory.com/reader/books/{epub_id}/packages/epub"
        )
        epub_data = response.json()
        encryption = AESEncryption(
            key = self._fix_key(epub_data["crypt_key"]),
            iv = self._fix_key(epub_data["crypt_iv"])
        )
        files = []
        for part in epub_data["spines"]:
            files.append(
                OnlineFile(
                    url = part["spine_url"],
                    extension = "epub",
                    encryption = encryption
                )
            )
        files_in_toc = {}
        for item in epub_data["toc"]["childrens"]: # Why is it "childrens"?
            files_in_toc[item["src"]] = item["name"]
        return EpubInParts(
            files,
            files_in_toc
        )



from grawlix.book import Book, BookData, SingleFile, ImageList, OnlineFile, HtmlFiles, EpubInParts
from grawlix.exceptions import GrawlixError, UnsupportedOutputFormat
from grawlix.logging import info

from .output_format import OutputFormat
from .acsm import Acsm
from .cbz import Cbz
from .epub import Epub

from typing import Callable, Iterable
from pathlib import Path
import os
import platform

async def download_book(book: Book, update_func: Callable, template: str) -> None:
    """
    Download and write book to disk

    :param book: Book to download
    """
    _, ext = os.path.splitext(template)
    ext = ext[1:]
    if ext in get_valid_extensions():
        output_format = find_output_format(book, ext)()
    else:
        output_format = get_default_format(book)
    location = format_output_location(book, output_format, template)
    if not book.overwrite and os.path.exists(location):
        info("Skipping - File already exists")
        return
    parent = Path(location).parent
    if not parent.exists():
        os.makedirs(parent)
    await output_format.download(book, location, update_func)
    await output_format.close()


def format_output_location(book: Book, output_format: OutputFormat, template: str) -> str:
    """
    Create path to output location of book

    :param book: Book to download
    :param output_format: Output format of book
    :param template: Template for output path
    :returns: Output path
    """
    values = { key: remove_unwanted_chars(value) for key, value in book.metadata.as_dict().items() }
    path = template.format(**values, ext = output_format.extension)
    return path


def remove_strings(input: str, strings: Iterable[str]) -> str:
    """
    Remove strings from input

    :param input: the string to remove strings from
    :param strings: the list of strings to remove from input
    :returns: input without strinfs
    """
    for c in strings:
        input = input.replace(c, "")
    return input


def remove_unwanted_chars(input: str) -> str:
    """
    Remove chars from string that are not supported in output path

    :param input: The string to remove chars from
    :returns: input without unsupported chars
    """
    if platform.system() == "Windows":
        return remove_strings(input, "<>:\"/\\|?*")
    else:
        return remove_strings(input, "/")


def get_default_format(book: Book) -> OutputFormat:
    """
    Get default output format for bookdata.
    Should only be used if no format was specified by the user

    :param book: Content of book
    :returns: OutputFormat object matching the default
    """
    bookdata = book.data
    if isinstance(bookdata, SingleFile):
        extension = bookdata.file.extension
    elif isinstance(bookdata, ImageList):
        extension = "cbz"
    elif isinstance(bookdata, HtmlFiles) or isinstance(bookdata, EpubInParts):
        extension = "epub"
    output_format = find_output_format(book, extension)
    return output_format()


def find_output_format(book: Book, extension: str) -> type[OutputFormat]:
    """
    Find a compatible output format

    :param book: Book to download
    :param extension: Extension of output file
    :returns: Compatible OutputFormat type
    :raises: UnsupportedOutputFormat if nothing is found
    """
    for output_format in get_output_formats():
        matches_extension = output_format.extension == extension
        supports_bookdata = type(book.data) in output_format.input_types
        if matches_extension and supports_bookdata:
            return output_format
    raise UnsupportedOutputFormat

def get_valid_extensions() -> list[str]:
    return [output_format.extension for output_format in get_output_formats()]


def get_output_formats() -> list[type[OutputFormat]]:
    """
    Get a list of all available output formats

    :returns: List of available output format classes
    """
    return [
        Acsm,
        Cbz,
        Epub,
    ]
